#include <Python.h>
#include <numpy/arrayobject.h>


static PyObject* hello_world (PyObject *self, PyObject *args) {
  printf("Hello_world\n");
  // C側で作成する関数のreturnはPy~~...となる
  Py_RETURN_NONE;
}

//2値化処理
static PyObject* binarizer(PyObject *self, PyObject *args)
{
	PyArrayObject *image_array;	    //The python objects to be extracted from the args
	int threshold;                  //The integer to be extracted from the args

    // Argument parsing, Object(image) and Integer
	if (!PyArg_ParseTuple(args, "Oi", &image_array, &threshold)) {
		fprintf(stderr, "invalid argument\n");
		Py_INCREF(Py_None);
		return Py_None;
	}

    // Get matrix dimensions
	int row_size = PyArray_DIM(image_array, 0);
	int columns_size = PyArray_DIM(image_array, 1);
	// Get the pointer of image data
    npy_intp* SnpArr = PyArray_STRIDES(image_array);


    // Get the new image area and pointer
    npy_intp dims[] = { row_size, columns_size, 1 };
	PyObject *new_image = PyArray_SimpleNew(2, dims, NPY_UBYTE);
	if (!new_image) {
		fprintf(stderr, "failed to allocate array\n");
		return Py_None;
	}
	PyArrayObject *new_image_array = (PyArrayObject *)new_image;

    // 2値化処理　ここではthresholdより大きい値を背景（0）としそうで無い場合をターゲット画素（1）に変換する
    // 画像データは2次元配列であるが1次元配列の体でポインタアクセスする。（バイトアクセス）
	for (int y = 0; y < row_size; y++) {
	    unsigned char const *src_array   = (unsigned char const *)image_array->data + (y) * columns_size;
	    unsigned char *dst_array   = (unsigned char *)new_image_array->data + (y) * columns_size;
	    for (int x = 0; x < columns_size; x++) {
	        if(src_array[x] > threshold){
	            dst_array[x] = 0;
	        }else{
	            dst_array[x] = 1;
	        }

	    }
	}
	return new_image;
}

static PyObject* summit_scanner(PyObject *self, PyObject *args)
{
	PyArrayObject *image_array;	    //The python objects to be extracted from the args
	 // Argument parsing, Object(image)
	if (!PyArg_ParseTuple(args, "O", &image_array)) {
		fprintf(stderr, "invalid argument\n");
		Py_INCREF(Py_None);
		return Py_None;
	}

    // Get matrix dimensions
	int row_size = PyArray_DIM(image_array, 0);
	int columns_size = PyArray_DIM(image_array, 1);
	// Get the pointer of image data
    npy_intp* SnpArr = PyArray_STRIDES(image_array);
//    頂点であると判断する15パターンのどれかに該当するか確認し、頂点であればその座標を記録する。
//    頂点は後に実行されるエッジトレースのスタート座標を示す。
//    b3 b2 b1
//    b4 b8 b0
//    b5 b6 b7
    int judge;
    int is_first = 1;
    int count = 0;
    int b8, b7, b6, b5, b4, b3, b2, b1, b0;
    unsigned char const *array_y_prev;
    unsigned char const *array_y;
    unsigned char const *array_y_next;

    PyObject* xy_list;
    PyObject* summit_list;
    summit_list = PyList_New(1);

	for (int y = 1; y < row_size-1; y++) {
	    array_y_prev = (unsigned char const *)image_array->data + (y-1) * columns_size;
	    array_y      = (unsigned char const *)image_array->data + (y) * columns_size;
	    array_y_next = (unsigned char const *)image_array->data + (y+1) * columns_size;
		for (int x = 1; x < columns_size-1; x++) {
            b8 = array_y[x];
            b7 = array_y_next[x + 1];
            b6 = array_y_next[x];
            b5 = array_y_next[x - 1];
            b4 = array_y[x - 1];
            b3 = array_y_prev[x - 1];
            b2 = array_y_prev[x];
            b1 = array_y_prev[x + 1];
            b0 = array_y[x + 1];
            judge = b8 & (~ (b1 | b2 | b3 | b4)) & (b0 | b5 | b6 | b7);
            count = count +  judge;
            if (judge){
                xy_list = PyList_New(2);
                PyList_SET_ITEM(xy_list, 0, PyLong_FromLong(x));
                PyList_SET_ITEM(xy_list, 1, PyLong_FromLong(y));
                if(is_first){
                    PyList_SET_ITEM(summit_list, 0, xy_list);
                    is_first = 0;
                } else {
                    PyList_Append(summit_list, xy_list);
                }
            }
		}
	}

    //printf("%d", judge);
	return summit_list;
}

static PyMethodDef myMethods[] = {
    { "hello_world", (PyCFunction)hello_world, METH_NOARGS, "my_module : hello_world"},
    { "binarizer", (PyCFunction)binarizer, METH_VARARGS, "my_module : binarizer" },
	{ "summit_scanner", (PyCFunction)summit_scanner, METH_VARARGS, "my_module : summit_scanner" },
	{ NULL }
};

static struct PyModuleDef my_module = {
	PyModuleDef_HEAD_INIT,
	"my_module",
	"Python3 C API Module",
	-1,
	myMethods
};

PyMODINIT_FUNC PyInit_my_module(void)
{
	import_array();
	return PyModule_Create(&my_module);
}